// -*- C++ -*-
/*!
 * @file  rtc_3DMapCalcAffineTransform.h
 * @brief Calculating the affine coefficients from the correspondence point data
 * @date  $Date$
 *
 * @author East Japan Accounting Center Co.,Ltd.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#ifndef RTC_3DMAPCALCAFFINETRANSFORM_H
#define RTC_3DMAPCALCAFFINETRANSFORM_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

//Add to code
#define _ON_INITIALIZE                 ("onInitialize execution")
#define _ON_FINALIZE                   ("onFinalize execution")
#define _ON_ACTIVATED                  ("onActivated execution")
#define _ON_DEACTIVATED                ("onDeactivated execution")
#define _ON_ABORTING                   ("onAborting execution")
#define _ON_RESET                      ("onReset execution")

#define _ERROR_FILE_NAME               ("ERROR: Input file is not found.")
#define _ERROR_FILE_CNT                ("ERROR: The number of input files is incorrect")
#define _ERROR_FILE_EXT                ("ERROR: The input file name is invalid. Please confirm the input file name.")
#define _ERROR_SAVE_FILE               ("ERROR: The directory with this file name does not exist")
#define _ERROR_SAVE_ALPNUM             ("ERROR: This file name contains something other than alphanumeric characters")
#define _ERROR_SAVE_FNLONG             ("ERROR: Filename is too long")
#define _STAGE_ERROR                   ("ERROR: Wrong value of the stage variable")
#define _ERROR_FILE_FORMAT             ("ERROR: Correspondence point coordinate data format is invalid.")
#define _ERROR_FILE_READ               ("ERROR: Failed to read the file")
#define _ERROR_FILE_SAVE               ("ERROR: Can't create output file of 3DMapCalcAffineTransform")
#define _ERROR_CONF_VAL                ("ERROR: The configuration value is out of range")
#define _ERROR_AXIS_RANGE              ("ERROR: This value of axis is out of range")
#define _ERROR_RESET                   ("ERROR: Error was detected. Please reset the 3DMapCalcAffineTransform RTC by the RT System Editor.")
#define _ERROR_SAVE_FILE_EXT           ("ERROR: The save file name is invalid. Please confirm the save file name.")

#define _3D_CALCAFFINE_SUCCESS            ("3DMapCalcAffineTransform RTC process was successful")
#define _3D_CALCAFFINE_FAILED             ("3DMapCalcAffineTransform RTC process was failed")

#define _3DCALCAFFINE_STAGE_INIT          (0)
#define _3DCALCAFFINE_STAGE_MIN           (_3DCALCAFFINE_STAGE_INIT)
#define _3DCALCAFFINE_STAGE1              (1)
#define _3DCALCAFFINE_STAGE2              (2)
#define _3DCALCAFFINE_STAGE3              (3)
#define ERROR_DETECTED                 _3DCALCAFFINE_STAGE3
#define _3DCALCAFFINE_STAGE_MAX           (_3DCALCAFFINE_STAGE3+1)

#define _IN_EXTENSION                  (".findpoint")  // lower case character only
#define _OUT_EXTENSION                 (".calcresult")  // lower case character only
#define _MAX_CHAR_CNT                  (32)
#define _EMPTY_STRING                  ("")
#define DEF_MAX_LENGTH                 (1024)
#define DEF_MID_LENGTH                 (512)
#define DEF_MIN_LENGTH                 (64)
#define DEF_GET_PATH_NG                (0)
#define DEF_GET_PATH_OK                (1)

#define _FIELDS_CNT                    (10)
#define _1ST                           (0)
#define _2ND                           (1)
#define _DELEMITER                     (' ')
#define _COMMENT_MARK                  ("#")
#define _PAGE                          (2)
#define _LSID                          (0)
#define _RSID                          (1)

//#define FILE_NAME_SEQ_COUNT            (4)
#define FILE_NAME_SEQ_COUNT            (1)
#define FILE_PATH_MODE_DIR             (0)
#define FILE_PATH_MODE_EXT             (1)
#define FILE_PATH_MODE_FIL             (2)
#define FILE_PATH_MODE_FNM             (3)
// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

//Add to code
#include "local_basic_type.h"
#include <mbstring.h>
#include <stdio.h>
#include <stdlib.h>
#include <Shlwapi.h>
#include <windows.h>
#include <algorithm>
/*!
 * @class rtc_3DMapCalcAffineTransform
 * @brief Calculating the affine coefficients from the correspondence point data
 *
 * This RT-component calculates affine coefficients and variance
 * data.
 *
 * The correspondence point data file path information.
 * The file name information to save the affine coefficient.
 *
 * -
 *
 * -
 *
 */
class rtc_3DMapCalcAffineTransform
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  rtc_3DMapCalcAffineTransform(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~rtc_3DMapCalcAffineTransform();

  // <rtc-template block="public_attribute">

  // </rtc-template>

  // <rtc-template block="public_operation">

  // </rtc-template>

  /***
   * Initialization of the 3DMapCalcAffineTransform RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry()
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * Finalize process the 3DMapCalcAffineTransform RTC
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   *
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   *
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the 3DMapCalcAffineTransform RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the 3DMapCalcAffineTransform RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of the 3DMapCalcAffineTransform RTC
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * onAborting the 3DMapCalcAffineTransform RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   *
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of the 3DMapCalcAffineTransform RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   * @pre -
   * @post -
   *
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);

  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   *
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   *
   *
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">

  // </rtc-template>

  // <rtc-template block="protected_operation">

  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::TimedString m_in_file_path;
  /*!
   * This port inputs the path information of the correspondence
   * point data file.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: one-cycle per second.
   */
  InPort<RTC::TimedString> m_in_file_pathIn;
  RTC::TimedString m_in_file_name;
  /*!
   * This port inputs the filename to save the affine
   * coefficient.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: one-cycle per second.
   */
  InPort<RTC::TimedString> m_in_file_nameIn;

  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">

  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">

  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">

  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">

  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">

  // </rtc-template>

  // <rtc-template block="private_operation">

  // </rtc-template>
  struct CALC_AFFINE
  {
      float_type affine_x;
      float_type affine_y;
      float_type var_z1;
      float_type mean_z1;
      float_type var_z2;
      float_type mean_z2;
      float_type depth;
  };

  struct MATCHED_POINT
  {
    uword_type m1;
    uword_type n1;
    float_type x1;
    float_type y1;
    float_type z1;
    uword_type m2;
    uword_type n2;
    float_type x2;
    float_type y2;
    float_type z2;
    sword_type diff_m1m2;
    sword_type diff_n1n2;
  };

  struct RANKING
  {
      sword_type diff_m1m2;
      sword_type diff_n1n2;
      sword_type cnt;
  };

  bool_type m_file_name_flag;
  bool_type m_save_file_flag;

  ubyte_type m_ubStageCtrl;
  ubyte_type m_ubStageCtrlBackUp;

  std::string m_inputfilename;
  std::string m_save_file_name;
  std::vector<MATCHED_POINT> m_read_match_point;
  std::vector<MATCHED_POINT> m_ranked_matched;
  /* Common */
  void nvdCalcOnInitialize(void);
  void nvdCalcOnFinalize(void);
  void nvdCalcOnActivated(void);
  void nvdCalcOnDeactivated(void);
  void nvdCalcOnAborting(void);
  void nvdCalcOnReset(void);
  void nvdClrMultipleFlags(void);
  void nvdClrMatchPointDatas(void);
  void nvdMapCalcAffineOnExecute(void);
  //void nvdClrFilenames(void);
  void nvdSetStageControl(ubyte_type a_ubStageCtrl);
  std::string sEditFilePath(std::string, uword_type);

  /* Stage1 */
  uword_type uw3DMapCalcAffineOnExecuteStage1(void);
  void nvdCheckofFileName(void);
  void nvdCheckofSaveFile(void);
  bool_type bCheckFileName(RTC::TimedString *);
  bool_type bCheckSaveFileName(RTC::TimedString *);
  bool_type bCheckFileExist(std::string file_name);
  bool_type bCheckExtension(std::string, std::string);
  bool_type bCheckDirExist(std::string folder_name);
  std::vector<std::string> split(std::string& , char_type );

  /* Stage2 */
  uword_type uw3DMapCalcAffineOnExecuteStage2(void);
  bool_type bReadFileFindpoint(RTC::TimedString *);
  bool_type bFindPointDataSplitter(std::ifstream *);
  bool_type bExtractRankingData();
  void nvdFactorCalculation(CALC_AFFINE&);
  void nvdAffineCalculation(CALC_AFFINE&);
  void nvdCountRankingData(std::vector<RANKING>& );
  void nvdDetermineRankingTopData(RANKING& , const std::vector<RANKING>& );
  void nvdCreateRankedMatchPointList(const RANKING&);
  bool_type bSaveResultFile(std::vector<CALC_AFFINE> *);
};


extern "C"
{
  DLL_EXPORT void rtc_3DMapCalcAffineTransformInit(RTC::Manager* manager);
};

#endif // RTC_3DMAPCALCAFFINETRANSFORM_H
