// -*- C++ -*-
/*!
 * @file  rtc_3DMapCalcAffineTransform.cpp
 * @brief Calculating the affine coefficients from the correspondence point data
 * @date $Date$
 *
 * @author East Japan Accounting Center Co.,Ltd.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#include "local_3DMapCalcAffineTransform.h"
#include "rtc_3DMapCalcAffineTransform.h"
#pragma comment(lib, "shlwapi.lib")

using namespace std;
using namespace pcl;
using namespace boost::accumulators;

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_3dmapcalcaffinetransform_spec[] =
  {
    "implementation_id", "rtc_3DMapCalcAffineTransform",
    "type_name",         "rtc_3DMapCalcAffineTransform",
    "description",       "Calculating the affine coefficients from the correspondence point data",
    "version",           "1.0",
    "vendor",            "EAC",
    "category",          "Image",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
rtc_3DMapCalcAffineTransform::rtc_3DMapCalcAffineTransform(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_in_file_pathIn("CorrespondencePointData", m_in_file_path),
    m_in_file_nameIn("SaveAffineFilename", m_in_file_name)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
rtc_3DMapCalcAffineTransform::~rtc_3DMapCalcAffineTransform()
{
}


/*!
 * Initialization of the 3DMapCalcAffineTransform RTC
 */
RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("CorrespondencePointData", m_in_file_pathIn);
  addInPort("SaveAffineFilename", m_in_file_nameIn);

  // Set OutPort buffer

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  nvdCalcOnInitialize();

  return RTC::RTC_OK;
}

/*!
 * Finalize process the 3DMapCalcAffineTransform RTC
 */

RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onFinalize()
{
  nvdCalcOnFinalize();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Activate the 3DMapCalcAffineTransform RTC
 */

RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onActivated(RTC::UniqueId ec_id)
{
  nvdCalcOnActivated();

  return RTC::RTC_OK;
}

/*!
 * Deactivate the 3DMapCalcAffineTransform RTC
 */

RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onDeactivated(RTC::UniqueId ec_id)
{
  nvdCalcOnDeactivated();

  return RTC::RTC_OK;
}

/*!
 * Periodic task of the 3DMapCalcAffineTransform RTC
 */

RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onExecute(RTC::UniqueId ec_id)
{
  nvdMapCalcAffineOnExecute();

  return RTC::RTC_OK;
}

/*!
 * onAborting the 3DMapCalcAffineTransform RTC
 */

RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onAborting(RTC::UniqueId ec_id)
{
  nvdCalcOnAborting();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Reset initialization of the 3DMapCalcAffineTransform RTC
 */

RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onReset(RTC::UniqueId ec_id)
{
  nvdCalcOnReset();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DMapCalcAffineTransform::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/***************************************************************************************************
*  Function        : nvdCalcOnInitialize
*  Description     : When onInitialize, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCalcOnInitialize(void)
{
  std::cout << _ON_INITIALIZE << std::endl;
}

/***************************************************************************************************
*  Function        : nvdCalcOnFinalize
*  Description     : When onFinalize, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCalcOnFinalize(void)
{
    std::cout << _ON_FINALIZE << std::endl;
}

/***************************************************************************************************
*  Function        : nvdCalcOnActivated
*  Description     : When onActivated, this function execute the initialization.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCalcOnActivated(void)
{
  std::cout << _ON_ACTIVATED << std::endl;

  m_ubStageCtrlBackUp = _3DCALCAFFINE_STAGE_INIT;
  m_ubStageCtrl = _3DCALCAFFINE_STAGE1;
  m_save_file_name = _EMPTY_STRING;
  m_inputfilename = _EMPTY_STRING;

  nvdClrMultipleFlags();
  nvdClrMatchPointDatas();
}

/***************************************************************************************************
*  Function        : nvdCalcOnDeactivated
*  Description     : When onDeactivated, this function execute the initialization.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCalcOnDeactivated(void)
{
  std::cout << _ON_DEACTIVATED << std::endl;
}

/***************************************************************************************************
*  Function        : nvdCalcOnAborting
*  Description     : When onAborting, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCalcOnAborting(void)
{
  std::cout << _ON_ABORTING << std::endl;
  std::cout << _ERROR_RESET << std::endl;
}

/***************************************************************************************************
*  Function        : nvdCalcOnReset
*  Description     : When onReset, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCalcOnReset(void)
{
  std::cout << _ON_RESET << std::endl;
}

/***************************************************************************************************
*  Function        : nvdMapFindOnExecute
*  Description     : This function execute the main process of 3DMapFindCorrespondence RTC.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdMapCalcAffineOnExecute(void)
{
  uword_type auwReturn = RESULT_NG;

  switch (m_ubStageCtrl)
  {
    case _3DCALCAFFINE_STAGE1:
      auwReturn = uw3DMapCalcAffineOnExecuteStage1();
      if (auwReturn == RESULT_OK) {
        nvdSetStageControl(_3DCALCAFFINE_STAGE2);
      }
      break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
    case _3DCALCAFFINE_STAGE2:
      auwReturn = uw3DMapCalcAffineOnExecuteStage2();
      if (auwReturn == RESULT_OK) {
        std::cout << _3D_CALCAFFINE_SUCCESS << std::endl;
        nvdSetStageControl(_3DCALCAFFINE_STAGE1);
      }
      else {
        nvdSetStageControl(ERROR_DETECTED);
      }
      break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
    case ERROR_DETECTED:
      if (m_ubStageCtrlBackUp != ERROR_DETECTED) {
        std::cout << _3D_CALCAFFINE_FAILED << std::endl;
      }
      break;
    //---------------------------------------------------------------------------------

    default:
      if (m_ubStageCtrlBackUp != m_ubStageCtrl) {
        std::cout << _STAGE_ERROR << std::endl;
      }
      break;
  }
  m_ubStageCtrlBackUp = m_ubStageCtrl;
}

/***************************************************************************************************
*  Function        : nvdSetStageControl
*  Description     : This function sets the member variable of m_ubStageCtrl.
*  Argument        : m_ubStageCtrl
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdSetStageControl(ubyte_type a_ubStageCtrl)
{
  if ((a_ubStageCtrl >= _3DCALCAFFINE_STAGE_MIN) && (a_ubStageCtrl < _3DCALCAFFINE_STAGE_MAX))
  {
    m_ubStageCtrl = a_ubStageCtrl;
  }
  else
  {
    m_ubStageCtrl = _3DCALCAFFINE_STAGE1;
  }
}

/***************************************************************************************************
*  Function        : uw3DMapFindCorrespondenceOnExecuteStage1
*  Description     : This function execute the main stage1 process of 3D Map Find Correspndence RTC.
*  Argument        : nothing special
*  Return          : information of normal end or abnormal end.
**************************************************************************************************/
uword_type rtc_3DMapCalcAffineTransform::uw3DMapCalcAffineOnExecuteStage1(void)
{
  uword_type auwReturn = RESULT_NG;

  if (!m_file_name_flag) {
    nvdCheckofFileName();
  }
  if (!m_save_file_flag) {
    nvdCheckofSaveFile();
  }
  if ((m_file_name_flag == TRUE) && (m_save_file_flag == TRUE))
  {
    nvdClrMultipleFlags();
    auwReturn = RESULT_OK;
  }

  return auwReturn;
}

/***************************************************************************************************
*  Function        : nvdCheckofFileName
*  Description     : This function checks the inputs image data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCheckofFileName(void)
{
  if (m_in_file_pathIn.isNew())
  {
    m_file_name_flag = FALSE;
    m_in_file_pathIn.read();
    if (bCheckFileName(&m_in_file_path))
    {
      nvdClrMatchPointDatas();
      m_file_name_flag = bReadFileFindpoint(&m_in_file_path);
      if (m_file_name_flag == TRUE)
      {
        /* code */
        for (auto &x : m_read_match_point) {
          x.diff_m1m2 = (x.m1 - x.m2);
          x.diff_n1n2 = (x.n1 - x.n2);
        }
      }
    }
  }
}

/***************************************************************************************************
*  Function        : bCheckFileName
*  Description     : This function checks the inputs filename.
*  Argument        : sequential file name
*  Return          : Returns true if the file exists and is correct, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bCheckFileName(RTC::TimedString *in_file_path)
{
  m_inputfilename = _EMPTY_STRING;

  std::string filefullpath = in_file_path->data;

  if (!bCheckFileExist(filefullpath))
  {
    std::cout << _ERROR_FILE_NAME << std::endl;
    return FALSE;
  }

  if (!bCheckExtension(filefullpath, _IN_EXTENSION))
  {
    std::cout << _ERROR_FILE_EXT << std::endl;
    return FALSE;
  }

  m_inputfilename = sEditFilePath(filefullpath, FILE_PATH_MODE_FNM);
  //}
  return TRUE;
}

/***************************************************************************************************
*  Function        : bCheckFileExist
*  Description     : This function checks the existence of a file.
*  Argument        : file name
*  Return          : Returns true if the file exists, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bCheckFileExist(std::string file_name)
{
  return (PathFileExists(file_name.c_str()));
  //return true;
}

/***************************************************************************************************
*  Function        : bCheckExtension
*  Description     : This function checks the extension.
*  Argument        : file path and target extension
*  Return          : Returns true if the file extension is correct, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bCheckExtension(std::string filepath, std::string targetext)
{
  std::string strEXT = sEditFilePath(filepath, FILE_PATH_MODE_EXT);
  /* Upper -> Lower */
  transform(strEXT.begin(), strEXT.end(), strEXT.begin(), tolower);

  if (strEXT == targetext)
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}

/***************************************************************************************************
*  Function        :bReadFileFindpoint
*  Description     :This function
*  Argument        :read file name (full path)
*  Return          :Returns true if the file exists, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bReadFileFindpoint(RTC::TimedString *in_file_path_p)
{
  bool_type abReturn = FALSE;

  std::string filefullpath = in_file_path_p->data;
  std::cout << "Read the file(" << filefullpath << ")" << std::endl;

  std::ifstream ifs(filefullpath);
  if (!ifs.fail())
  {
    abReturn = bFindPointDataSplitter(&ifs);
  }
  else
  {
    /* File read error. */
    std::cout << _ERROR_FILE_READ << std::endl;
  }

  return abReturn;
}

/***************************************************************************************************
*  Function        : bFindPointDataSplitter
*  Description     : This function splits the trajectory file.
*  Argument        : input file stream and image buffer
*  Return          : Returns true if the file can be split, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bFindPointDataSplitter(std::ifstream *ifs)
{
  std::string str_line;
  std::vector<std::string> strFields;
  MATCHED_POINT matched_buff;

  while (std::getline(*ifs, str_line))
  {
    /* Comment line goto next. */
    if (str_line.substr(0, 1) == _COMMENT_MARK) continue;

    strFields = split(str_line, _DELEMITER);
    if (strFields.size() == _FIELDS_CNT)
    {
      try
      {
        matched_buff.m1 = stoi(strFields[0]);
        matched_buff.n1 = stoi(strFields[1]);
        matched_buff.x1 = stof(strFields[2]);
        matched_buff.y1 = stof(strFields[3]);
        matched_buff.z1 = stof(strFields[4]);
        matched_buff.m2 = stoi(strFields[5]);
        matched_buff.n2 = stoi(strFields[6]);
        matched_buff.x2 = stof(strFields[7]);
        matched_buff.y2 = stof(strFields[8]);
        matched_buff.z2 = stof(strFields[9]);

        m_read_match_point.push_back(matched_buff);
      }
      catch (...)
      {
        /* Format error. */
        std::cout << _ERROR_FILE_FORMAT << std::endl;
        return FALSE;
      }
    }
    else
    {
      /* Format error. */
      std::cout << _ERROR_FILE_FORMAT << std::endl;
      return FALSE;
    }
  }
  if (m_read_match_point.empty()) {
    /* Format error. */
    std::cout << _ERROR_FILE_FORMAT << std::endl;
    return FALSE;
  }
  return TRUE;
}

/***************************************************************************************************
*  Function        : nvdCheckofSaveFile
*  Description     : This function checks the inputs image data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCheckofSaveFile(void)
{
  if (m_in_file_nameIn.isNew())
  {
    m_save_file_name = _EMPTY_STRING;
    m_save_file_flag = FALSE;
    m_in_file_nameIn.read();
    if (bCheckSaveFileName(&m_in_file_name))
    {
      m_save_file_name = m_in_file_name.data;
      m_save_file_flag = TRUE;
    }
  }
}

/***************************************************************************************************
*  Function        : bCheckSaveFileName
*  Description     : This function checks the save filename.
*  Argument        : save file name (full path)
*  Return          : Returns true if the file can be saved, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bCheckSaveFileName(RTC::TimedString *idp_savefilename)
{
  // TimedString -> string
  std::string strTemp = idp_savefilename->data;
  std::string strFilename = sEditFilePath(strTemp, FILE_PATH_MODE_FIL);

  // alphanumeric check
  if ( !std::all_of(strFilename.cbegin(), strFilename.cend(), _ismbcalnum) )
  {
    std::cout << _ERROR_SAVE_ALPNUM << std::endl;
    return FALSE;
  }

  // character count check
  if (strFilename.length() > _MAX_CHAR_CNT)
  {
    std::cout << _ERROR_SAVE_FNLONG << std::endl;
    return FALSE;
  }

  // directory existence check
  if ( !bCheckDirExist(sEditFilePath(strTemp, FILE_PATH_MODE_DIR)) )
  {
    std::cout << _ERROR_SAVE_FILE << std::endl;
    return FALSE;
  }

  if (!bCheckExtension(strTemp, _OUT_EXTENSION))
  {
    std::cout << _ERROR_SAVE_FILE_EXT << std::endl;
    return FALSE;
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : nvdClrMultipleFlags
*  Description     : This function clears multiple flags.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdClrMultipleFlags(void)
{
  m_file_name_flag = FALSE;
  m_save_file_flag = FALSE;
}

/***************************************************************************************************
*  Function        : nvdClrMatchPointDatas
*  Description     : This function clears multiple Image
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdClrMatchPointDatas(void)
{
  m_read_match_point.clear();
  m_ranked_matched.clear();
  m_read_match_point.shrink_to_fit();
  m_ranked_matched.shrink_to_fit();
}

/***************************************************************************************************
*  Function        : bCheckDirExist
*  Description     : This function checks the existence of a directory.
*  Argument        : folder name
*  Return          : Returns true if the directory exists, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bCheckDirExist(std::string folder_name)
{
  return (PathIsDirectory(folder_name.c_str()));
  //return true;

}

/***************************************************************************************************
*  Function        : uw3DMapCalcAffineOnExecuteStage2
*  Description     : This function execute the main stage2 process
*  Argument        : nothing special
*  Return          : information of normal end or abnormal end.
**************************************************************************************************/
uword_type rtc_3DMapCalcAffineTransform::uw3DMapCalcAffineOnExecuteStage2(void)
{
  uword_type auwReturn = RESULT_NG;

  std::vector<CALC_AFFINE> vec_calced;
  CALC_AFFINE calced_buff;

  if (m_read_match_point.empty()) {
    return auwReturn;
  }

  /* extract RankingData */
  if (!bExtractRankingData())
  {
    /* error */
    return auwReturn;
  }

  nvdFactorCalculation(calced_buff);
  nvdAffineCalculation(calced_buff);

  vec_calced.push_back(calced_buff);

  /* File write */
  if (!bSaveResultFile(&vec_calced))
  {
    std::cout << _ERROR_FILE_SAVE << std::endl;
    return auwReturn;
  }
  auwReturn = RESULT_OK;

  return auwReturn;
}

/***************************************************************************************************
*  Function        :bExtractRankingData
*  Description     :This function extract ranked data
*  Argument        :nothing special
*  Return          :information of normal end or abnormal end.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bExtractRankingData()
{
  std::vector<RANKING> ranked;
  RANKING top;
  top.diff_n1n2 = 0;
  top.diff_m1m2 = 0;
  top.cnt = 0;

  nvdCountRankingData(ranked);
  nvdDetermineRankingTopData(top, ranked);
  nvdCreateRankedMatchPointList(top);

  if (m_ranked_matched.empty()) {
    return FALSE;
  }
  return TRUE;
}

/***************************************************************************************************
*  Function        :nvdCountRankingData
*  Description     :This function ranking data count
*  Argument        :ranked pointer
*  Return          :nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCountRankingData(std::vector<RANKING>& ranked_list)
{
  for (auto x : m_read_match_point) {
    uword_type flg = 0;

    for (auto &l : ranked_list) {
      if ((x.diff_m1m2 == l.diff_m1m2) &&
           (x.diff_n1n2 == l.diff_n1n2))
      {
        l.cnt++;
        flg = 1;
        break;
      }
    }
    if (flg == 0) {
      RANKING ranked_buff;
      ranked_buff.diff_m1m2 = x.diff_m1m2;
      ranked_buff.diff_n1n2 = x.diff_n1n2;
      ranked_buff.cnt = 1;
      ranked_list.push_back(ranked_buff);
    }
  }
}

/***************************************************************************************************
*  Function        :nvdDetermineRankingTopData
*  Description     :This function Judging the top ranking
*  Argument        :top pointer,ranked reference
*  Return          :nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdDetermineRankingTopData(RANKING& top_buff, const std::vector<RANKING>& ranked_list)
{
  for (auto x : ranked_list) {
    if (top_buff.cnt < x.cnt) {
      top_buff.diff_m1m2 = x.diff_m1m2;
      top_buff.diff_n1n2 = x.diff_n1n2;
      top_buff.cnt = x.cnt;
    }
  }
}

/***************************************************************************************************
*  Function        :nvdCreateRankedMatchPointList
*  Description     :This function create ranked match point data list
*  Argument        :top reference
*  Return          :nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdCreateRankedMatchPointList(const RANKING& top_buff)
{
  MATCHED_POINT matched_buff;

  if (top_buff.cnt <= 1) {
    for (auto x : m_read_match_point) {
      matched_buff.x1 = x.x1;
      matched_buff.y1 = x.y1;
      matched_buff.z1 = x.z1;
      matched_buff.x2 = x.x2;
      matched_buff.y2 = x.y2;
      matched_buff.z2 = x.z2;
      m_ranked_matched.push_back(matched_buff);
    }
  }
  else {
    for (auto x : m_read_match_point) {
      if ((x.diff_m1m2 == top_buff.diff_m1m2) &&
          (x.diff_n1n2 == top_buff.diff_n1n2))
      {
        matched_buff.x1 = x.x1;
        matched_buff.y1 = x.y1;
        matched_buff.z1 = x.z1;
        matched_buff.x2 = x.x2;
        matched_buff.y2 = x.y2;
        matched_buff.z2 = x.z2;
        m_ranked_matched.push_back(matched_buff);
      }
    }
  }
}

/***************************************************************************************************
*  Function        :nvdFactorCalculation
*  Description     :This function calculation Z1/Z2 Variance,Mean,Depth.
*  Argument        :calculation data buff reference
*  Return          :nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdFactorCalculation(CALC_AFFINE& calced_buff)
{
  accumulator_set<double, stats<tag::mean, tag::variance> > acc1;
  accumulator_set<double, stats<tag::mean, tag::variance> > acc2;

  /* Variance calculation */
  for (auto x : m_ranked_matched)
  {
    acc1(x.z1);
    acc2(x.z2);
  }

  calced_buff.var_z1 = variance(acc1);
  calced_buff.var_z2 = variance(acc2);

  cout << "var1:    " << calced_buff.var_z1 << std::endl;
  cout << "var2:    " << calced_buff.var_z2 << std::endl;

  /* mean calculation */
  calced_buff.mean_z1 = mean(acc1);
  calced_buff.mean_z2 = mean(acc2);

  cout << "mean1:   " << calced_buff.mean_z1 << std::endl;
  cout << "mean2:   " << calced_buff.mean_z2 << std::endl;

  /* depth Coefficient calculation*/
  if (calced_buff.var_z2 == 0) {
    calced_buff.depth = 0.0;
  }
  else {
    calced_buff.depth = sqrt(calced_buff.var_z1 / calced_buff.var_z2);
  }
  cout << "depth:   " << calced_buff.depth << std::endl;
}

/***************************************************************************************************
*  Function        :nvdAffineCalculation
*  Description     :This function calculation Affine Coefficient.
*  Argument        :calculation data buff reference
*  Return          :nothing special
**************************************************************************************************/
void rtc_3DMapCalcAffineTransform::nvdAffineCalculation(CALC_AFFINE& calced_buff)
{
  std::vector<pcl::PointXYZ> fcorres1st, fcorres2nd;
  pcl::PointXYZ XYZ_buff1;
  pcl::PointXYZ XYZ_buff2;

  for (auto x : m_ranked_matched)
  {
    //XYZ_buff1.x = x.x1;
    //XYZ_buff1.y = x.y1;
    XYZ_buff1.x = 0.0;
    XYZ_buff1.y = 0.0;
    XYZ_buff1.z = 0.0;
    XYZ_buff2.x = (x.x1 - x.x2);
    XYZ_buff2.y = (x.y1 - x.y2);
    XYZ_buff2.z = 0.0;
    fcorres1st.push_back(XYZ_buff1);
    fcorres2nd.push_back(XYZ_buff2);
  }

  Eigen::Affine3f testdata = computeTransform(fcorres2nd, fcorres1st);
  //Eigen::Affine3f testdata = computeTransform(fcorres1st, fcorres2nd);

  /* calc affine Coefficient */
  calced_buff.affine_x = testdata(0, 3);
  calced_buff.affine_y = testdata(1, 3);

  std::cout << "affine_x: " << calced_buff.affine_x << std::endl;
  std::cout << "affine_y: " << calced_buff.affine_y << std::endl;

}

/***************************************************************************************************
*  Function        : sEditFilePath
*  Description     : This function edit the file path.
*  Argument        : file name and process mode
*  Return          : Returns a string of directory path or filename or extension.
**************************************************************************************************/
std::string rtc_3DMapCalcAffineTransform::sEditFilePath(std::string filename, uword_type mode)
{
  char_type strPathName[DEF_MAX_LENGTH];

  char_type strFullPath[DEF_MAX_LENGTH];
  char_type strDrive[DEF_MIN_LENGTH];
  char_type strDir[DEF_MAX_LENGTH];
  char_type strFile[DEF_MID_LENGTH];
  char_type strExt[DEF_MIN_LENGTH];

  memset(strFullPath, 0x00, sizeof(strFullPath));
  memset(strDrive, 0x00, sizeof(strDrive));
  memset(strDir, 0x00, sizeof(strDir));
  memset(strFile, 0x00, sizeof(strFile));
  memset(strExt, 0x00, sizeof(strExt));
  memset(strPathName, 0x00, sizeof(strPathName));

  //string -> char[]
  filename.copy(strFullPath, filename.size());
  _splitpath(strFullPath, strDrive, strDir, strFile, strExt);

  switch (mode){
    case FILE_PATH_MODE_DIR:
      sprintf(strPathName, "%s%s", strDrive, strDir);
      break;

    case FILE_PATH_MODE_EXT:
      sprintf(strPathName, "%s", strExt);
      break;

    case FILE_PATH_MODE_FIL:
      sprintf(strPathName, "%s", strFile);
      break;

    case FILE_PATH_MODE_FNM:
    default:
      sprintf(strPathName, "%s%s", strFile, strExt);
    break;
  }

  return strPathName;
}

/***************************************************************************************************
*  Function        : bSaveResultFile
*  Description     : This function saves the matching trajectory in a file.
*  Argument        : matched data
*  Return          : Returns true if the matched data file can be saved, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapCalcAffineTransform::bSaveResultFile(std::vector<CALC_AFFINE> *calced)
{
  FILE* fp;

  std::cout << "Save the file(" << m_save_file_name << ")" << std::endl;

  if (m_save_file_name == _EMPTY_STRING){
    return FALSE;
  }

  fp = fopen(m_save_file_name.c_str(), "wb");
  if (fp == NULL){
    return FALSE;
  }

  fprintf(fp, "# Created by 3DMapCalcAffineTransform\n");
  fprintf(fp, "# input filename : %s\n", m_inputfilename.c_str());
  fprintf(fp, "# Data       : \n");

  for (auto itr = calced->begin(); itr != calced->end(); ++itr)
  {
    fprintf(fp, "%.6f %.6f %.6f %.6f %.6f %.6f %.6f\n"
              , itr->affine_x, itr->affine_y
              , itr->var_z1, itr->mean_z1
              , itr->var_z2, itr->mean_z2
              , itr->depth);
  }

  fclose(fp);

  return TRUE;
}

/***************************************************************************************************
*  Function        : split
*  Description     : This function splits the string.
*  Argument        : Line string and delimiter
*  Return          : Returns the string after split.
**************************************************************************************************/
std::vector<std::string> rtc_3DMapCalcAffineTransform::split(std::string& str_line, char_type delimiter)
{
  std::istringstream stream(str_line);
  std::string field;
  std::vector<std::string> result;

  while (getline(stream, field, delimiter))
  {
    result.push_back(field);
  }
  return result;
}

/***************************************************************************************************
*  Function        :
*  Description     :
*  Argument        :
*  Return          :
**************************************************************************************************/
Eigen::Affine3f computeTransform(const std::vector<pcl::PointXYZ>& prev_points, const std::vector<pcl::PointXYZ>& curr_points)
{
  pcl::TransformationFromCorrespondences tfc;

  Eigen::Matrix<float, 3, Eigen::Dynamic> w(3, prev_points.size());
  Eigen::Matrix<float, 3, Eigen::Dynamic> m(3, curr_points.size());

  for (size_t i = 0; i < prev_points.size(); ++i) {
    w(0, i) = prev_points[i].x;
    w(1, i) = prev_points[i].y;
    w(2, i) = prev_points[i].z;
    m(0, i) = curr_points[i].x;
    m(1, i) = curr_points[i].y;
    m(2, i) = curr_points[i].z;

    tfc.add(m.col(i), w.col(i)); // tfc will return a matrix T such that T * curr = prev.
  }

  return tfc.getTransformation();
}

/***************************************************************************************************
*  Function        :
*  Description     :
*  Argument        :
*  Return          :
**************************************************************************************************/

extern "C"
{

  void rtc_3DMapCalcAffineTransformInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_3dmapcalcaffinetransform_spec);
    manager->registerFactory(profile,
                             RTC::Create<rtc_3DMapCalcAffineTransform>,
                             RTC::Delete<rtc_3DMapCalcAffineTransform>);
  }

};
